#include "mublas_devel.h"

void
mublasCreate (mublasHandle_t* mublashandle) {
    int driverVersion = 0;
	int runtimeVersion = 0;
	cudaDriverGetVersion(&driverVersion);
	cudaRuntimeGetVersion(&runtimeVersion);

	// Device Properties
	cudaDeviceProp devprop;  
	cudaGetDeviceProperties (&devprop, 0);

	mublashandle->cc = devprop.major * 100 + devprop.minor * 10;

	switch (mublashandle->cc) {
		case 100:
			mublashandle->W_M_pmax = 24;
			mublashandle->B_M_pmax = 8;
			mublashandle->R_unit = 256;
			mublashandle->S_unit = 512;
			mublashandle->W_gran = 2;
			mublashandle->Cores = 8;
			mublashandle->DPCores = 1;
			mublashandle->B_M_optmin = B_M_min_Fermi;
			break;
		case 110:
			mublashandle->W_M_pmax = 24;
			mublashandle->B_M_pmax = 8;
			mublashandle->R_unit = 256;
			mublashandle->S_unit = 512;
			mublashandle->W_gran = 2;
			mublashandle->Cores = 8;
			mublashandle->DPCores = 1;
			mublashandle->B_M_optmin = B_M_min_Fermi;
			break;
		case 120:
			mublashandle->W_M_pmax = 32;
			mublashandle->B_M_pmax = 8;
			mublashandle->R_unit = 512;
			mublashandle->S_unit = 512;
			mublashandle->W_gran = 2;
			mublashandle->Cores = 8;
			mublashandle->DPCores = 1;
			mublashandle->B_M_optmin = B_M_min_Fermi;
			break;
		case 130:
			mublashandle->W_M_pmax = 32;
			mublashandle->B_M_pmax = 8;
			mublashandle->R_unit = 512;
			mublashandle->S_unit = 512;
			mublashandle->W_gran = 2;
			mublashandle->Cores = 8;
			mublashandle->DPCores = 1;
			mublashandle->B_M_optmin = B_M_min_Fermi;
			break;
		case 200:
			mublashandle->W_M_pmax = 48;
			mublashandle->B_M_pmax = 8;
			mublashandle->R_unit = 64;
			mublashandle->S_unit = 128;
			mublashandle->W_gran = 2;
			mublashandle->Cores = 32;
			mublashandle->DPCores = 16;
			mublashandle->B_M_optmin = B_M_min_Fermi;
			break;
		case 210:
			mublashandle->W_M_pmax = 48;
			mublashandle->B_M_pmax = 8;
			mublashandle->R_unit = 64;
			mublashandle->S_unit = 128;
			mublashandle->W_gran = 2;
			mublashandle->Cores = 48;
			mublashandle->DPCores = 6;
			mublashandle->B_M_optmin = B_M_min_Fermi;
			break;
		case 300:
			mublashandle->W_M_pmax = 64;
			mublashandle->B_M_pmax = 16;
			mublashandle->R_unit = 256;
			mublashandle->S_unit = 256;
			mublashandle->W_gran = 4;
			mublashandle->Cores = 192;
			mublashandle->DPCores = 64;
			mublashandle->B_M_optmin = B_M_min_Kepler;
			break;
		case 350:
			mublashandle->W_M_pmax = 64;
			mublashandle->B_M_pmax = 16;
			mublashandle->R_unit = 256;
			mublashandle->S_unit = 256;
			mublashandle->W_gran = 4;
			mublashandle->Cores = 192;
			mublashandle->DPCores = 64;
			mublashandle->B_M_optmin = B_M_min_Kepler;
			break;
		case 500:
			mublashandle->W_M_pmax = 64;
			mublashandle->B_M_pmax = 32;
			mublashandle->R_unit = 256;
			mublashandle->S_unit = 256;
			mublashandle->W_gran = 4;
			mublashandle->Cores = 128;
			mublashandle->DPCores = 4;
			mublashandle->B_M_optmin = B_M_min_Maxwell;
			break;
		case 520:
			mublashandle->W_M_pmax = 64;
			mublashandle->B_M_pmax = 32;
			mublashandle->R_unit = 256;
			mublashandle->S_unit = 256;
			mublashandle->W_gran = 4;
			mublashandle->Cores = 128;
			mublashandle->DPCores = 4;
			mublashandle->B_M_optmin = B_M_min_Maxwell;
			break;
		default: // CC350 
			mublashandle->W_M_pmax = 64;
			mublashandle->B_M_pmax = 16;
			mublashandle->R_unit = 256;
			mublashandle->S_unit = 256;
			mublashandle->W_gran = 4;
			mublashandle->Cores = 192;
			mublashandle->DPCores = 64;
			mublashandle->B_M_optmin = B_M_min_Kepler;
			break;
	}

	mublashandle->devprop = devprop;
	mublashandle->driverVersion = driverVersion;
	mublashandle->runtimeVersion = runtimeVersion;
	mublashandle->Tx_B = 0;
	mublashandle->Ty_B = 0;
	mublashandle->Tx_B_static = 0;
	mublashandle->Ty_B_static = 0;
	mublashandle->S_B = 0;
	mublashandle->B_D = 0;
	mublashandle->BO = 0;
	mublashandle->WO = 0;
	mublashandle->kernelnumber = 0;
}
