#ifndef MUBLAS_DEVEL_H
#define MUBLAS_DEVEL_H

#include <stdio.h>
#define __STDC_FORMAT_MACROS
#include <inttypes.h>
#include <cuda_runtime_api.h>
#include <cublas_v2.h>

#define min(a, b) ((a) < (b) ? (a) : (b))
#define max(a, b) ((a) > (b) ? (a) : (b))

#include "../include/mublas.h"

// ***********************************************
// Threads per warp (warp size)
#define T_W 32
// B_M: 1-8 (Fermi), 1-16 (Kepler), 1-32 (Maxwell)
#define B_M_min_Fermi 1
#define B_M_min_Kepler 3
#define B_M_min_Maxwell 1
// Num of threads per block
#define Th_B_min 128
#define Th_B_max_GEMVN 1024
#define Th_B_max_GEMVT 512
// ***********************************************

// Ceiling and floor function (Excel compatible) -----------------------
__host__ __inline__ int32_t ceilingEx (const float n, const int32_t c) {
	return (int32_t)ceil(n/c)*c;
}

__host__ __inline__ int32_t floorEx (const float n, const int32_t c) {
	return (int32_t)floor(n/c)*c;
}
// ---------------------------------------------------------------------

extern __host__ void get_occupancy ( mublasHandle_t* mh);
extern __host__ void kernel_configuration_phase_1 (mublasHandle_t* mh);
extern __host__ void kernel_configuration_phase_2 (mublasHandle_t* mh);

#ifdef PREC_S
#include "prec_s.h"
#endif
#ifdef PREC_D
#include "prec_d.h"
#endif
#ifdef PREC_C
#include "prec_c.h"
#endif
#ifdef PREC_Z
#include "prec_z.h"
#endif
#ifdef PREC_W
#include "prec_w.h"
#endif
#ifdef PREC_U
#include "prec_u.h"
#endif
#ifdef PREC_B
#include "prec_b.h"
#endif
#ifdef PREC_V
#include "prec_v.h"
#endif
#ifdef PREC_DB
#include "prec_db.h"
#endif

#endif
