/* ***********************************************
 
	GEMV for CUDA 
	update: Mar. 25, 2015

   *********************************************** */   

#include "mublas_devel.h"

// Read Only Data Cache
#define USE_NC_MAT_A_GEMVN //default yes
//#define USE_NC_VEC_X_GEMVN //default no
//#define USE_NC_MAT_A_GEMVT //default no
#define USE_NC_VEC_X_GEMVT //default yes

#if defined (USE_NC_MAT_A_GEMVN)
#define NC_MAT_A_GEMVN * __restrict__
#else
#define NC_MAT_A_GEMVN *
#endif
#if defined (USE_NC_VEC_X_GEMVN)
#define NC_VEC_X_GEMVN * __restrict__
#else
#define NC_VEC_X_GEMVN *
#endif
#if defined (USE_NC_MAT_A_GEMVT)
#define NC_MAT_A_GEMVT * __restrict__
#else
#define NC_MAT_A_GEMVT *
#endif
#if defined (USE_NC_VEC_X_GEMVT)
#define NC_VEC_X_GEMVT * __restrict__
#else
#define NC_VEC_X_GEMVT *
#endif
 
// ----- functions determining Tx_B and Ty_B -----
__host__ void get_B_D_gemvn ( mublasHandle_t* mh) {
	mh->B_D = (int32_t) ceil ((float)mh->m / (float)(mh->N_T * mh->Tx_B));
}

__host__ void get_B_D_gemvt ( mublasHandle_t* mh) {
	mh->B_D = (int32_t) ceil ((float)mh->n / (float)(mh->Tx_B / T_W));
}

__host__ void get_S_B_gemvn ( mublasHandle_t* mh) {
	mh->S_B = ceilingEx ((float)(mh->Tx_B * mh->Ty_B * mh->N_T * mh->N_one), mh->S_unit);
}

__host__ void get_S_B_gemvt ( mublasHandle_t* mh) {
	if (mh->cc >= 300)
		mh->S_B = ceilingEx ((float)(mh->Tx_B * mh->N_one), mh->S_unit);
	else
		mh->S_B = ceilingEx ((float)(mh->Tx_B * mh->N_one + T_W * mh->N_one * mh->N_T), mh->S_unit);
}

__host__ void get_Occupancy_gemv ( mublasHandle_t* mh) {
	const int32_t R_T = mh->funcatb.numRegs;
	const int32_t M_D = mh->devprop.multiProcessorCount;
	const int32_t S_B_pmax = mh->devprop.sharedMemPerBlock;
	const int32_t R_M = mh->devprop.regsPerBlock;
	const int32_t R_W = ceilingEx ((float)(R_T * T_W), mh->R_unit); // Usage of registers per warp
	const int32_t W_M_rmax = min (floorEx ((float)R_M / (float)R_W, mh->W_gran), mh->W_M_pmax); // Max warps per multiprocessor
	const int32_t W_B = (int32_t) ceil ((float)(mh->Tx_B * mh->Ty_B) / (float)T_W); // Warps per block
	const int32_t B_M_rmax = (int32_t) floor ((float)W_M_rmax / (float)W_B); // Blocks per multiprocessor (limited by register)
	int32_t B_M_smax; // Blocks per multiprocessor (limited by shared memory)
	if (mh->S_B == 0) {
		mh->B_M_max = min (B_M_rmax, mh->B_M_pmax); // Blocks per multiprocessor
	} else {
		B_M_smax = (int32_t) floor ((float)S_B_pmax / (float)mh->S_B);
		mh->B_M_max = min (min (B_M_rmax, B_M_smax), mh->B_M_pmax); // Blocks per multiprocessor
	}
	const int32_t B_D_max = mh->B_M_max * M_D;
	mh->BO = (float)mh->B_D / (float)ceilingEx ((float)mh->B_D, B_D_max);
	int32_t W_M_max = W_B * mh->B_M_max; // Warps per multiprocessor
	mh->WO = (float)W_M_max / (float)mh->W_M_pmax; // Warp Occupancy
}

/* ***********************************************
	MUBLAS-SET-PARAM for N-kernel
   *********************************************** */   
__host__ void get_Param_gemvn_1 (mublasHandle_t* mh, int64_t lda, const TYPE0 *Ag, const TYPE0 *Xg, const TYPE0 *Yg) {
	int32_t Tx_B_step;
	int32_t B_D_min;

// --- 1. alignment check and set initial value --------------------------------------
	if      (VEC4) mh->N_T = 4;
	else if (VEC2) mh->N_T = 2;
	else           mh->N_T = 1;

	// alignment check
	if ((((long)Ag%(128/sizeof(TYPE0)))%16 == 0) && 
		(((long)Xg%(128/sizeof(TYPE0)))%16 == 0) && 
		(((long)Yg%(128/sizeof(TYPE0)))%16 == 0) && (lda%mh->N_T == 0))
		mh->USEvector = 1;
	else 
		mh->USEvector = 0;

	// Tx_B_step, B_M_min, B_D_min
	if (mh->cc >= 300)
		Tx_B_step = 128 / (mh->N_one * mh->N_T); // 128bytes/16bytes(vec)=8thread
	else
		Tx_B_step = 16; // half-warp
	mh->B_M_min = mh->B_M_optmin;
	B_D_min = mh->devprop.multiProcessorCount * mh->B_M_min;

// --- 2. determine B_M_min, Tx_B_min ------------------------------------------------
	mh->Tx_B = T_W; // default 
	get_B_D_gemvn (mh);
	// B_M_min
	while (mh->B_D < B_D_min && mh->B_M_min > 1) {
		mh->B_M_min -= 1;
		get_B_D_gemvn (mh);
	}
	// Tx_B_min
	while (mh->B_D < B_D_min && mh->Tx_B > Tx_B_step) {
		mh->Tx_B -= Tx_B_step;
		get_B_D_gemvn (mh);
	}
	mh->Tx_B_min = mh->Tx_B;
}

__host__ void get_Param_gemvn_2 (mublasHandle_t* mh) {
	int32_t Tx_B_tmp;
	int32_t Tx_B_step;
	int32_t Ty_B_min;
	int32_t Ty_B_max;
	int32_t Ty_B_tmp;

	if (mh->cc >= 300) Tx_B_step = 128 / (mh->N_one * mh->N_T); // 128bytes/16bytes(vec)=8thread
	else Tx_B_step = 16; // half-warp

// --- 3. determine Tx_B, Ty_B ------------------------------------------------
	if (mh->Tx_B_static != 0 && mh->Ty_B_static != 0) {
		Tx_B_tmp = mh->Tx_B_static;
		Ty_B_tmp = mh->Ty_B_static;
	}
	else {
		Tx_B_tmp = mh->Tx_B_min;
		Ty_B_tmp = (int32_t)ceil (Th_B_min / (float)mh->Tx_B_min);
		float BO_tmp = 0.;
		float WO_tmp = 0.;
		// --- Tx_B ------------------------------------------------
		for (mh->Tx_B = mh->Tx_B_min; mh->Tx_B <= Th_B_max_GEMVN; mh->Tx_B += Tx_B_step) { 
			get_B_D_gemvn (mh);
			// --- Ty_B -----------------------------------------
			Ty_B_min = (int32_t)ceil (Th_B_min / (float)mh->Tx_B);
			Ty_B_max = max ( Ty_B_min, (int32_t)floor ((float)Th_B_max_GEMVN / (float)mh->Tx_B));
			for (mh->Ty_B = Ty_B_max; mh->Ty_B >= Ty_B_min; mh->Ty_B--) { 
				get_S_B_gemvn (mh);
				get_Occupancy_gemv (mh);
				if (mh->B_M_max >= mh->B_M_min && mh->BO >= BO_tmp) {
					if (mh->BO == 1.) {
						if (mh->WO >= WO_tmp) {
							BO_tmp = mh->BO;
							WO_tmp = mh->WO;
							Tx_B_tmp = mh->Tx_B;
							Ty_B_tmp = mh->Ty_B;
						}
					} else {
						BO_tmp = mh->BO;
						WO_tmp = mh->WO;
						Tx_B_tmp = mh->Tx_B;
						Ty_B_tmp = mh->Ty_B;
					}
				}
			}
		}
	}

// --- 4. flush ------------------------------------------------
	mh->Tx_B = Tx_B_tmp;
	mh->Ty_B = Ty_B_tmp;
	get_B_D_gemvn (mh);
	get_S_B_gemvn (mh);
	get_Occupancy_gemv (mh);
}

/* ***********************************************
	MUBLAS-SET-PARAM for T-kernel
   *********************************************** */   
__host__ void get_Param_gemvt_1 (mublasHandle_t* mh, int64_t lda, const TYPE0 *Ag, const TYPE0 *Xg, const TYPE0 *Yg) {
	int32_t B_D_min;

// --- 1. set initial value ------------------------------------------------
	if      (VEC4) mh->N_T = 4;
	else if (VEC2) mh->N_T = 2;
	else           mh->N_T = 1;

	// alignment check
	if ((((long)Ag%(128/sizeof(TYPE0)))%16 == 0) && 
		(((long)Xg%(128/sizeof(TYPE0)))%16 == 0) && 
		(((long)Yg%(128/sizeof(TYPE0)))%16 == 0) && (lda%mh->N_T == 0))
		mh->USEvector = 1;
	else 
		mh->USEvector = 0;

// --- 2. determine B_M_min, Tx_B_min ------------------------------------------------
	// default
	mh->B_M_min = mh->B_M_optmin;
	B_D_min = mh->devprop.multiProcessorCount * mh->B_M_min;
	mh->Tx_B = T_W;
	get_B_D_gemvt (mh);

	// B_M_min
	while (mh->B_D < B_D_min && mh->B_M_min > 1) {
		mh->B_M_min -= 1;
		get_B_D_gemvt (mh);
	}
	// Tx_B_min
	mh->Tx_B_min = mh->Tx_B;
}

__host__ void get_Param_gemvt_2 (mublasHandle_t* mh) {
	int32_t Tx_B_tmp;
	int32_t Tx_B_step;
	int32_t B_D_min;

	Tx_B_step = T_W;
	mh->Ty_B = 1;

// --- 3. determine Tx_B ------------------------------------------------
	if (mh->Tx_B_static != 0) {
		Tx_B_tmp = mh->Tx_B_static;
	}
	else {
		Tx_B_tmp = mh->Tx_B_min;
		B_D_min = mh->devprop.multiProcessorCount * mh->B_M_min;
		float BO_tmp = 0.;
		float WO_tmp = 0.;
		// --- Tx_B ------------------------------------------------
		for (mh->Tx_B = Th_B_min; mh->Tx_B <= Th_B_max_GEMVT; mh->Tx_B += Tx_B_step) { 
			get_B_D_gemvt (mh);
			if (mh->B_D >= B_D_min) {
				get_S_B_gemvt (mh);
				get_Occupancy_gemv (mh);
				if (mh->B_M_max >= mh->B_M_min && mh->BO >= BO_tmp) {
					if (mh->BO == 1.) {
						if (mh->WO >= WO_tmp) {
							BO_tmp = mh->BO;
							WO_tmp = mh->WO;
							Tx_B_tmp = mh->Tx_B;
						}
					} else {
						BO_tmp = mh->BO;
						WO_tmp = mh->WO;
						Tx_B_tmp = mh->Tx_B;
					}
				}
			}
		}
	}

// --- 4. flush ------------------------------------------------
	mh->Tx_B = Tx_B_tmp;
	get_B_D_gemvt (mh);
	get_S_B_gemvt (mh);
	get_Occupancy_gemv (mh);
}

/* ***********************************************
	N-kernel (s1)
   *********************************************** */   
__global__ void MUBLAS_GEMV_N_DEVICE_s1 (
	const int32_t m,
	const int32_t n,
	const TYPE0 alpha,
	const TYPE0 NC_MAT_A_GEMVN Ag,
	const int32_t lda,
	const TYPE0 NC_VEC_X_GEMVN Xg,
	const TYPE0 beta,
	TYPE0 *Yg
) {
#if VEC1
	const int32_t Tx_B = blockDim.x;
	const int32_t Ty_B = blockDim.y;
	const int32_t T_B = Tx_B * Ty_B;
	const int32_t iTx = threadIdx.x;
	const int32_t iTy = threadIdx.y;
	const int32_t iTh = iTy * Tx_B + iTx;
	const int32_t rowid = (blockIdx.x * Tx_B + iTx);
	int32_t i, j, k;
	extern __shared__ TYPE1 Xs1[];
	TYPE1 Tr1 = ZERO1;

	// main
	for (i = 0; i < n/T_B; i++) {
		k = i*T_B;
		Xs1[iTh] = Xg[k + iTh];
		__syncthreads ();
		if (rowid < m) {
			for (j = iTy; j < T_B; j+= Ty_B) 
				Tr1 = MAD1 (Xs1[j], Ag[(j+k) * lda + rowid], Tr1);
		}
		__syncthreads ();
	}

	// last one block
	k = i*T_B;
	if (k + iTh < n) Xs1[iTh] = Xg[k + iTh];
	__syncthreads ();
	if (rowid < m) {
		for (j = k + iTy; j < n; j += Ty_B) {
			if (j < n) Tr1 = MAD1 (Xs1[j-k], Ag[j * lda + rowid], Tr1);
		}
	}
	__syncthreads ();

	// reduction
	Xs1[iTh] = Tr1;
	__syncthreads ();
	if (iTy == 0) {
		for (i = 1; i < Ty_B; i++) 
			Tr1 = ADD1 (Tr1, Xs1[Tx_B * i + iTx]);
		if (rowid < m) Yg[rowid] = MAD1 (alpha, Tr1, MUL1 (beta, Yg[rowid]));
	}
#endif
}


/* ***********************************************
	N-kernel (v2)
   *********************************************** */   
__global__ void MUBLAS_GEMV_N_DEVICE_v2 (
	const int32_t m,
	const int32_t n,
	const TYPE0 alpha,
	const TYPE0 NC_MAT_A_GEMVN Ag,
	const int32_t lda,
	const TYPE0 NC_VEC_X_GEMVN Xg,
	const TYPE0 beta,
	TYPE0 *Yg
) {
#if VEC2
	const int32_t N_T = 2;
	const int32_t Tx_B = blockDim.x;
	const int32_t Ty_B = blockDim.y;
	const int32_t T_B = Tx_B * Ty_B;
	const int32_t iTx = threadIdx.x;
	const int32_t iTy = threadIdx.y;
	const int32_t iTh = iTy * Tx_B + iTx;
	const int32_t rowid = (blockIdx.x * Tx_B + iTx) * N_T;
	int32_t i, j, k;
	extern __shared__ TYPE2 Xs2[];
	TYPE2 Tr2 = ZERO2;

	// main
	for (i = 0; i < n/(T_B*N_T); i++) {
		k = i*T_B*N_T;
		Xs2[iTh] = (TYPE2&)Xg[k + iTh * N_T];
		__syncthreads ();
		if (rowid < m) {
			for (j = iTy*N_T; j < T_B*N_T; j+= Ty_B*N_T) {
				Tr2 = MAD122 ((TYPE1&)Xs2[j/N_T].x, (TYPE2&)Ag[(j+k+0) * lda + rowid], Tr2);
				#if defined PREC_C || defined PREC_B || defined PREC_DB
				Tr2 = MAD122 ((TYPE1&)Xs2[j/N_T].z, (TYPE2&)Ag[(j+k+1) * lda + rowid], Tr2);
				#else
				Tr2 = MAD122 ((TYPE1&)Xs2[j/N_T].y, (TYPE2&)Ag[(j+k+1) * lda + rowid], Tr2);
				#endif
			}
		}
		__syncthreads ();
	}

	// last one block
	k = i*T_B*N_T;
	if (n % N_T == 0) {
		if (k + iTh * N_T + 0 < n) Xs2[iTh] = (TYPE2&)Xg[k + iTh * N_T];
	}
	else {
		if (k + iTh * N_T + 0 < n) (TYPE1&)Xs2[iTh].x = Xg[k + iTh * N_T + 0];
		#if defined PREC_C || defined PREC_B || defined PREC_DB
		if (k + iTh * N_T + 1 < n) (TYPE1&)Xs2[iTh].z = Xg[k + iTh * N_T + 1];
		#else
		if (k + iTh * N_T + 1 < n) (TYPE1&)Xs2[iTh].y = Xg[k + iTh * N_T + 1];
		#endif
	}
	__syncthreads ();
	if (rowid < m) {
		for (j = k + iTy * N_T; j < n; j += Ty_B*N_T) {
			if (j + 0 < n) Tr2 = MAD122 ((TYPE1&)Xs2[(j-k)/N_T].x, (TYPE2&)Ag[(j+0) * lda + rowid], Tr2);
			#if defined PREC_C || defined PREC_B || defined PREC_DB
			if (j + 1 < n) Tr2 = MAD122 ((TYPE1&)Xs2[(j-k)/N_T].z, (TYPE2&)Ag[(j+1) * lda + rowid], Tr2);
			#else
			if (j + 1 < n) Tr2 = MAD122 ((TYPE1&)Xs2[(j-k)/N_T].y, (TYPE2&)Ag[(j+1) * lda + rowid], Tr2);
			#endif
		}
	}
	__syncthreads ();

	// reduction
	Xs2[iTh] = Tr2;
	__syncthreads ();
	if (iTy == 0) {
		for (i = 1; i < Ty_B; i++) 
			Tr2 = ADD2 (Tr2, Xs2[Tx_B * i + iTx]);
		if (m % N_T == 0) {
			if (rowid < m) (TYPE2&) Yg[rowid] = MAD122 (alpha, Tr2, MUL12 (beta, (TYPE2&)Yg[rowid]));
		} else {
			if (rowid+0 < m) Yg[rowid+0] = MAD1 (alpha, (TYPE1&)Tr2.x, MUL1 (beta, Yg[rowid+0]));
			#if defined PREC_C || defined PREC_B || defined PREC_DB
			if (rowid+1 < m) Yg[rowid+1] = MAD1 (alpha, (TYPE1&)Tr2.z, MUL1 (beta, Yg[rowid+1]));
			#else
			if (rowid+1 < m) Yg[rowid+1] = MAD1 (alpha, (TYPE1&)Tr2.y, MUL1 (beta, Yg[rowid+1]));
			#endif
		}
	}
#endif
}

/* ***********************************************
	N-kernel (v4)
   *********************************************** */   
__global__ void MUBLAS_GEMV_N_DEVICE_v4 (
	const int32_t m,
	const int32_t n,
	const TYPE0 alpha,
	const TYPE0 NC_MAT_A_GEMVN Ag,
	const int32_t lda,
	const TYPE0 NC_VEC_X_GEMVN Xg,
	const TYPE0 beta,
	TYPE0 *Yg
) {
#if VEC4
	const int32_t N_T = 4;
	const int32_t Tx_B = blockDim.x;
	const int32_t Ty_B = blockDim.y;
	const int32_t T_B = Tx_B * Ty_B;
	const int32_t iTx = threadIdx.x;
	const int32_t iTy = threadIdx.y;
	const int32_t iTh = iTy * Tx_B + iTx;
	const int32_t rowid = (blockIdx.x * Tx_B + iTx) * N_T;
	int32_t i, j, k;
	extern __shared__ TYPE4 Xs4[];
	TYPE4 Tr4 = ZERO4;

	// main
	for (i = 0; i < n/(T_B*N_T); i++) {
		k = i*T_B*N_T;
		Xs4[iTh] = (TYPE4&)Xg[k + iTh * N_T];
		__syncthreads ();
		if (rowid < m) {
			for (j = iTy*N_T; j < T_B*N_T; j+= Ty_B*N_T) {
				Tr4 = MAD144 (Xs4[j/N_T].x, (TYPE4&)Ag[(j+k+0) * lda + rowid], Tr4);
				Tr4 = MAD144 (Xs4[j/N_T].y, (TYPE4&)Ag[(j+k+1) * lda + rowid], Tr4);
				Tr4 = MAD144 (Xs4[j/N_T].z, (TYPE4&)Ag[(j+k+2) * lda + rowid], Tr4);
				Tr4 = MAD144 (Xs4[j/N_T].w, (TYPE4&)Ag[(j+k+3) * lda + rowid], Tr4);
			}
		}
		__syncthreads ();
	}

	// last one block
	k = i*T_B*N_T;
	if (n % N_T == 0) {
		if (k + iTh * N_T < n) Xs4[iTh] = (TYPE4&)Xg[k + iTh * N_T];
	}
	else {
		if (k + iTh * N_T + 0 < n) Xs4[iTh].x = Xg[k + iTh * N_T + 0];
		if (k + iTh * N_T + 1 < n) Xs4[iTh].y = Xg[k + iTh * N_T + 1];
		if (k + iTh * N_T + 2 < n) Xs4[iTh].z = Xg[k + iTh * N_T + 2];
		if (k + iTh * N_T + 3 < n) Xs4[iTh].w = Xg[k + iTh * N_T + 3];
	}
	__syncthreads ();
	if (rowid < m) {
		for (j = k + iTy * N_T; j < n; j += Ty_B*N_T) {
			if (j + 0 < n) Tr4 = MAD144 (Xs4[(j-k)/N_T].x, (TYPE4&)Ag[(j+0) * lda + rowid], Tr4);
			if (j + 1 < n) Tr4 = MAD144 (Xs4[(j-k)/N_T].y, (TYPE4&)Ag[(j+1) * lda + rowid], Tr4);
			if (j + 2 < n) Tr4 = MAD144 (Xs4[(j-k)/N_T].z, (TYPE4&)Ag[(j+2) * lda + rowid], Tr4);
			if (j + 3 < n) Tr4 = MAD144 (Xs4[(j-k)/N_T].w, (TYPE4&)Ag[(j+3) * lda + rowid], Tr4);
		}
	}
	__syncthreads ();

	// reduction
	Xs4[iTh] = Tr4;
	__syncthreads ();
	if (iTy == 0) {
		for (i = 1; i < Ty_B; i++) 
			Tr4 = ADD4 (Tr4, Xs4[Tx_B * i + iTx]);
		if (m % N_T == 0) {
			if (rowid < m) (TYPE4&) Yg[rowid] = MAD144 (alpha, Tr4, MUL14 (beta, (TYPE4&)Yg[rowid]));
		} else {
			if (rowid+0 < m) Yg[rowid+0] = MAD1 (alpha, Tr4.x, MUL1 (beta, Yg[rowid+0]));
			if (rowid+1 < m) Yg[rowid+1] = MAD1 (alpha, Tr4.y, MUL1 (beta, Yg[rowid+1]));
			if (rowid+2 < m) Yg[rowid+2] = MAD1 (alpha, Tr4.z, MUL1 (beta, Yg[rowid+2]));
			if (rowid+3 < m) Yg[rowid+3] = MAD1 (alpha, Tr4.w, MUL1 (beta, Yg[rowid+3]));
		}
	}
#endif
}

/* ***********************************************
	N-kernel (s2)
   *********************************************** */   
__global__ void MUBLAS_GEMV_N_DEVICE_s2 (
	const int32_t m,
	const int32_t n,
	const TYPE0 alpha,
	const TYPE0 NC_MAT_A_GEMVN Ag,
	const int32_t lda,
	const TYPE0 NC_VEC_X_GEMVN Xg,
	const TYPE0 beta,
	TYPE0 *Yg
) {
	const int32_t N_T = 2;
	const int32_t Tx_B = blockDim.x;
	const int32_t Ty_B = blockDim.y;
	const int32_t T_B = Tx_B * Ty_B;
	const int32_t iTx = threadIdx.x;
	const int32_t iTy = threadIdx.y;
	const int32_t iTh = iTy * Tx_B + iTx;
	const int32_t rowid = (blockIdx.x * Tx_B + iTx) * N_T;
	int32_t i, j, k, l;
	extern __shared__ TYPE0 Xs1[];
	TYPE0 Tr1[2];

	#pragma unroll
	for (l = 0; l < N_T; l++) Tr1[l] = ZERO1;

	// main
	for (i = 0; i < n/T_B; i++) {
		k = i*T_B;
		Xs1[iTh] = Xg[k + iTh];
		__syncthreads ();
		if (rowid < m) {
			for (j = iTy; j < T_B; j+= Ty_B) {
				#pragma unroll
				for (l = 0; l < N_T; l++) Tr1[l] = MAD1 (Xs1[j], Ag[(j+k) * lda + rowid+l], Tr1[l]);
			}
		}
		__syncthreads ();
	}

	// last one block
	k = i*T_B;
	if (k + iTh < n)
		Xs1[iTh] = Xg[k + iTh];
	__syncthreads ();
	if (rowid < m) {
		for (j = k + iTy; j < n; j += Ty_B) {
			#pragma unroll
			for (l = 0; l < N_T; l++) Tr1[l] = MAD1 (Xs1[j-k], Ag[j * lda + rowid+l], Tr1[l]);
		}
	}
	__syncthreads ();

	// reduction
	#pragma unroll
	for (l = 0; l < N_T; l++) Xs1[iTh*N_T+l] = Tr1[l];
	__syncthreads ();
	if (iTy == 0) {
		for (i = 1; i < Ty_B; i++) {
			#pragma unroll
			for (l = 0; l < N_T; l++) Tr1[l] = ADD1 (Tr1[l], Xs1[(Tx_B * i + iTx)*N_T+l]);
		}
		#pragma unroll
		for (l = 0; l < N_T; l++) {
			if (rowid+l < m) Yg[rowid+l] = MAD1 (alpha, Tr1[l], MUL1 (beta, Yg[rowid+l]));
		}
	}
}

/* ***********************************************
	N-kernel (s4)
   *********************************************** */   
__global__ void MUBLAS_GEMV_N_DEVICE_s4 (
	const int32_t m,
	const int32_t n,
	const TYPE0 alpha,
	const TYPE0 NC_MAT_A_GEMVN Ag,
	const int32_t lda,
	const TYPE0 NC_VEC_X_GEMVN Xg,
	const TYPE0 beta,
	TYPE0 *Yg
) {
	const int32_t N_T = 4;
	const int32_t Tx_B = blockDim.x;
	const int32_t Ty_B = blockDim.y;
	const int32_t T_B = Tx_B * Ty_B;
	const int32_t iTx = threadIdx.x;
	const int32_t iTy = threadIdx.y;
	const int32_t iTh = iTy * Tx_B + iTx;
	const int32_t rowid = (blockIdx.x * Tx_B + iTx) * N_T;
	int32_t i, j, k, l;
	extern __shared__ TYPE0 Xs1[];
	TYPE0 Tr1[4];

	#pragma unroll
	for (l = 0; l < N_T; l++) Tr1[l] = ZERO1;

	// main
	for (i = 0; i < n/T_B; i++) {
		k = i*T_B;
		Xs1[iTh] = Xg[k + iTh];
		__syncthreads ();
		if (rowid < m) {
			for (j = iTy; j < T_B; j+= Ty_B) {
				#pragma unroll
				for (l = 0; l < N_T; l++) Tr1[l] = MAD1 (Xs1[j], Ag[(j+k) * lda + rowid+l], Tr1[l]);
			}
		}
		__syncthreads ();
	}

	// last one block
	k = i*T_B;
	if (k + iTh < n)
		Xs1[iTh] = Xg[k + iTh];
	__syncthreads ();
	if (rowid < m) {
		for (j = k + iTy; j < n; j += Ty_B) {
			#pragma unroll
			for (l = 0; l < N_T; l++) Tr1[l] = MAD1 (Xs1[j-k], Ag[j * lda + rowid+l], Tr1[l]);
		}
	}
	__syncthreads ();

	// reduction
	#pragma unroll
	for (l = 0; l < N_T; l++) Xs1[iTh*N_T+l] = Tr1[l];
	__syncthreads ();
	if (iTy == 0) {
		for (i = 1; i < Ty_B; i++) {
			#pragma unroll
			for (l = 0; l < N_T; l++) Tr1[l] = ADD1 (Tr1[l], Xs1[(Tx_B * i + iTx)*N_T+l]);
		}
		#pragma unroll
		for (l = 0; l < N_T; l++) {
			if (rowid+l < m) Yg[rowid+l] = MAD1 (alpha, Tr1[l], MUL1 (beta, Yg[rowid+l]));
		}
	}
}

/* ***********************************************
	T-kernel (s1)
   *********************************************** */   
__global__ void MUBLAS_GEMV_T_DEVICE_s1 (
	const int32_t m,
	const int32_t n,
	const TYPE0 alpha,
	const TYPE0 NC_MAT_A_GEMVT Ag,
	const int32_t lda,
	const TYPE0 NC_VEC_X_GEMVT Xg,
	const TYPE0 beta,
	TYPE0 *Yg
) {
#if VEC1
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
#else
	int32_t tx = threadIdx.x;
	extern __shared__ TYPE1 Ts1[];
#endif
	int32_t tid = blockDim.x * blockIdx.x + threadIdx.x;
	int32_t rowid = tid / T_W;
	int32_t lane  = tid % T_W;
	int32_t i;
	__shared__ TYPE1 Xs1[T_W];
	TYPE1 Tr1;

	Tr1 = ZERO1;

	for (i = lane; i < m; i += T_W) {
		Xs1[lane] = Xg[i];
		Tr1 = MAD1 (Ag[min(n-1, rowid) * lda + i], Xs1[lane], Tr1);
	}

	// reduction
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
	#pragma unroll
	for (i = T_W/2; i > 0; i >>= 1) 
		shuffleAdd (Tr1, i);
#else
	Ts1[tx] = Tr1;
	if (lane < 16) Ts1[tx] = ADD1 (Ts1[tx+16], Ts1[tx]);
	if (lane <  8) Ts1[tx] = ADD1 (Ts1[tx+ 8], Ts1[tx]);
	if (lane <  4) Ts1[tx] = ADD1 (Ts1[tx+ 4], Ts1[tx]);
	if (lane <  2) Ts1[tx] = ADD1 (Ts1[tx+ 2], Ts1[tx]);
	if (lane <  1) Tr1 = ADD1 (Ts1[tx+ 1], Ts1[tx]);
#endif

	if (lane == 0) {
		if (rowid < n)
			Yg[rowid] = MAD1 (alpha, Tr1, MUL1 (beta, Yg[rowid]));
	}
#endif
}

/* ***********************************************
	T-kernel (v2)
   *********************************************** */   
__global__ void MUBLAS_GEMV_T_DEVICE_v2 (
	const int32_t m,
	const int32_t n,
	const TYPE0 alpha,
	const TYPE0 NC_MAT_A_GEMVT Ag,
	const int32_t lda,
	const TYPE0 NC_VEC_X_GEMVT Xg,
	const TYPE0 beta,
	TYPE0 *Yg
) {
#if VEC2
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
#else
	int32_t tx = threadIdx.x;
	extern __shared__ TYPE1 Ts1[];
#endif
	int32_t tid = blockDim.x * blockIdx.x + threadIdx.x;
	int32_t rowid = (tid / T_W);
	int32_t lane  = tid % T_W;
	int32_t i, j, k;
	const int32_t N_T = 2;
	__shared__ TYPE2 Xs2[T_W];
	TYPE1 Tr1;
	TYPE2 Tr2;

	Tr2 = ZERO2;
	Tr1 = ZERO1;

	for (i = 0; i < m / (T_W*N_T); i++) {
		j = i*T_W*N_T + lane*N_T;
		Xs2[lane] = (TYPE2&)Xg[j];
		Tr2 = MAD2 ((TYPE2&)Ag[min(n-1, rowid) * lda + j], Xs2[lane], Tr2);
	}
#if defined PREC_C || defined PREC_B || defined PREC_DB
	Tr1 = ADD1 ((TYPE1&)Tr2.x, (TYPE1&)Tr2.z);
#else
	Tr1 = ADD1 (Tr2.x, Tr2.y);
#endif

	if (m % (T_W*N_T) != 0) { 
		for (k = i*T_W*N_T + lane; k < m; k += T_W) 
			Tr1 = MAD1 (Ag[min(n-1, rowid) * lda + k], Xg[k], Tr1);
	}

	// reduction
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
	#pragma unroll
	for (i = T_W/2; i > 0; i >>= 1) 
		shuffleAdd (Tr1, i);
#else
	Ts1[tx] = Tr1;
	if (lane < 16) Ts1[tx] = ADD1 (Ts1[tx+16], Ts1[tx]);
	if (lane <  8) Ts1[tx] = ADD1 (Ts1[tx+ 8], Ts1[tx]);
	if (lane <  4) Ts1[tx] = ADD1 (Ts1[tx+ 4], Ts1[tx]);
	if (lane <  2) Ts1[tx] = ADD1 (Ts1[tx+ 2], Ts1[tx]);
	if (lane <  1) Tr1 = ADD1 (Ts1[tx+ 1], Ts1[tx]);
#endif

	if (lane == 0) {
		if (rowid < n)
			Yg[rowid] = MAD1 (alpha, Tr1, MUL1 (beta, Yg[rowid]));
	}
#endif
}

/* ***********************************************
	T-kernel (v4)
   *********************************************** */   
__global__ void MUBLAS_GEMV_T_DEVICE_v4 (
	const int32_t m,
	const int32_t n,
	const TYPE0 alpha,
	const TYPE0 NC_MAT_A_GEMVT Ag,
	const int32_t lda,
	const TYPE0 NC_VEC_X_GEMVT Xg,
	const TYPE0 beta,
	TYPE0 *Yg
) {
#if VEC4
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
#else
	int32_t tx = threadIdx.x;
	extern __shared__ TYPE1 Ts1[];
#endif
	int32_t tid = blockDim.x * blockIdx.x + threadIdx.x;
	int32_t rowid = (tid / T_W);
	int32_t lane  = tid % T_W;
	int32_t i, j, k;
	const int32_t N_T = 4;
	__shared__ TYPE4 Xs4[T_W];
	TYPE1 Tr1;
	TYPE4 Tr4;

	Tr4 = ZERO4;
	Tr1 = ZERO1;

	for (i = 0; i < m / (T_W*N_T); i++) {
		j = i*T_W*N_T + lane*N_T;
		Xs4[lane] = (TYPE4&)Xg[j];
		Tr4 = MAD4 ((TYPE4&)Ag[min(n-1, rowid) * lda + j], Xs4[lane], Tr4);
	}
	Tr1 = ADD1 (ADD1 (Tr4.x, Tr4.y), ADD1 (Tr4.z, Tr4.w));

	if (m % (T_W*N_T) != 0) { 
		for (k = i*T_W*N_T + lane; k < m; k += T_W) 
			Tr1 = MAD1 (Ag[min(n-1, rowid) * lda + k], Xg[k], Tr1);
	}

	// reduction
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
	#pragma unroll
	for (i = T_W/2; i > 0; i >>= 1) 
		shuffleAdd (Tr1, i);
#else
	Ts1[tx] = Tr1;
	if (lane < 16) Ts1[tx] = ADD1 (Ts1[tx+16], Ts1[tx]);
	if (lane <  8) Ts1[tx] = ADD1 (Ts1[tx+ 8], Ts1[tx]);
	if (lane <  4) Ts1[tx] = ADD1 (Ts1[tx+ 4], Ts1[tx]);
	if (lane <  2) Ts1[tx] = ADD1 (Ts1[tx+ 2], Ts1[tx]);
	if (lane <  1) Tr1 = ADD1 (Ts1[tx+ 1], Ts1[tx]);
#endif

	if (lane == 0) {
		if (rowid < n)
			Yg[rowid] = MAD1 (alpha, Tr1, MUL1 (beta, Yg[rowid]));
	}
#endif
}

/* ***********************************************
	Host Interface
   *********************************************** */   

__host__ int32_t
MUBLAS_GEMV (
	mublasHandle_t* mh,	
	char trans,
	const int32_t m,
	const int32_t n,
	const TYPE0 *alpha,
	const TYPE0 *a,
	const int32_t lda,
	const TYPE0 *x,
	const int32_t incx,
	const TYPE0 *beta,
	TYPE0 *y,
	const int32_t incy
) {

    if (m == 0 || n == 0 || (*alpha == ZERO0 && *beta == ONE0 )) 
		return 0;
	#if CUBLAS_SCAL_HOST != NULL
	if (*alpha == ZERO0) {
		cublasHandle_t ch;
		cublasCreate (&ch);
		cublasSetPointerMode(ch, CUBLAS_POINTER_MODE_HOST);
		CUBLAS_SCAL (ch, m, beta, y, incy);
		return 0;
	}
	#endif

	mh->N_one = sizeof (TYPE0);
	mh->m = m;
	mh->n = n;

// ----------------------------
	if (trans == 'N' || trans == 'n') {
		if (incx == 1 && incy == 1) {

			get_Param_gemvn_1 (mh, lda, a, x, y);

			if (mh->N_T == 4) { 
				if (mh->USEvector == 1) { // 4x4 vector version
					mh->kernelnumber = 140;
					cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_N_DEVICE_v4);
					get_Param_gemvn_2 (mh);
					dim3 threads (mh->Tx_B, mh->Ty_B);
					dim3 grid (mh->B_D);
					cudaFuncSetCacheConfig (MUBLAS_GEMV_N_DEVICE_v4, cudaFuncCachePreferShared);
					MUBLAS_GEMV_N_DEVICE_v4 <<< grid, threads, mh->S_B >>> (m, n, *alpha, a, lda, x, *beta, y);
				} else { // for misaligned, scalar version
					mh->kernelnumber = 141;
					cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_N_DEVICE_s4);
					get_Param_gemvn_2 (mh);
					dim3 threads (mh->Tx_B, mh->Ty_B);
					dim3 grid (mh->B_D);
					cudaFuncSetCacheConfig (MUBLAS_GEMV_N_DEVICE_s4, cudaFuncCachePreferShared);
					MUBLAS_GEMV_N_DEVICE_s4 <<< grid, threads, mh->S_B >>> (m, n, *alpha, a, lda, x, *beta, y);
				}
			}
			else if (mh->N_T == 2) { 
				if (mh->USEvector == 1) { // 2x2 vector version
					mh->kernelnumber = 120;
					cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_N_DEVICE_v2);
					get_Param_gemvn_2 (mh);
					dim3 threads (mh->Tx_B, mh->Ty_B);
					dim3 grid (mh->B_D);
					cudaFuncSetCacheConfig (MUBLAS_GEMV_N_DEVICE_v2, cudaFuncCachePreferShared);
					MUBLAS_GEMV_N_DEVICE_v2 <<< grid, threads, mh->S_B >>> (m, n, *alpha, a, lda, x, *beta, y);
				} else { // for misaligned, scalar version
					mh->kernelnumber = 121;
					cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_N_DEVICE_s2);
					get_Param_gemvn_2 (mh);
					dim3 threads (mh->Tx_B, mh->Ty_B);
					dim3 grid (mh->B_D);
					cudaFuncSetCacheConfig (MUBLAS_GEMV_N_DEVICE_s2, cudaFuncCachePreferShared);
					MUBLAS_GEMV_N_DEVICE_s2 <<< grid, threads, mh->S_B >>> (m, n, *alpha, a, lda, x, *beta, y);
				}
			} else { // scalar version
				mh->kernelnumber = 110;
				cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_N_DEVICE_s1);
				get_Param_gemvn_2 (mh);
				dim3 threads (mh->Tx_B, mh->Ty_B);
				dim3 grid (mh->B_D);
				cudaFuncSetCacheConfig (MUBLAS_GEMV_N_DEVICE_s1, cudaFuncCachePreferShared);
				MUBLAS_GEMV_N_DEVICE_s1 <<< grid, threads, mh->S_B >>> (m, n, *alpha, a, lda, x, *beta, y);
			}
		}
		else { // INCX,INCY != 1
			#if defined (PREC_S) || defined (PREC_C) || defined (PREC_D) || defined (PREC_Z)
			mh->kernelnumber = 8100;
			cublasHandle_t ch;
			cublasCreate (&ch);
			cublasSetPointerMode(ch, CUBLAS_POINTER_MODE_HOST);
			cublasOperation_t opAcublas = CUBLAS_OP_N;
			CUBLAS_GEMV (ch, opAcublas, m, n, alpha, a, lda, x, incx, beta, y, incy);
			#else
			printf ("mublas error: not implemented");
			#endif
		}
	}
	else {
		if (incx == 1 && incy == 1) {
			
			get_Param_gemvt_1 (mh, lda, a, x, y);

			if (mh->N_T == 4) { 
				if (mh->USEvector == 1) { // 4 vector version
					mh->kernelnumber = 240;
					cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_T_DEVICE_v4);
					get_Param_gemvt_2 (mh);
					dim3 threads (mh->Tx_B, mh->Ty_B);
					dim3 grid (mh->B_D);
					cudaFuncSetCacheConfig (MUBLAS_GEMV_T_DEVICE_v4, cudaFuncCachePreferL1);
					if (mh->cc >= 300) MUBLAS_GEMV_T_DEVICE_v4 <<< grid, threads >>> (m, n, *alpha, a, lda, x, *beta, y);
					else MUBLAS_GEMV_T_DEVICE_v4 <<< grid, threads, mh->S_B >>> (m, n, *alpha, a, lda, x, *beta, y);
				} else { // for misaligned, scalar version
					mh->kernelnumber = 210;
					cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_T_DEVICE_s1);
					get_Param_gemvt_2 (mh);
					dim3 threads (mh->Tx_B, mh->Ty_B);
					dim3 grid (mh->B_D);
					cudaFuncSetCacheConfig (MUBLAS_GEMV_T_DEVICE_s1, cudaFuncCachePreferL1);
					if (mh->cc >= 300) MUBLAS_GEMV_T_DEVICE_s1 <<< grid, threads >>> (m, n, *alpha, a, lda, x, *beta, y);
					else MUBLAS_GEMV_T_DEVICE_s1 <<< grid, threads, mh->S_B >>> (m, n, *alpha, a, lda, x, *beta, y);
				}
			} else if (mh->N_T == 2) {
				if (mh->USEvector == 1) { // 2 vector version
					mh->kernelnumber = 220;
					cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_T_DEVICE_v2);
					get_Param_gemvt_2 (mh);
					dim3 threads (mh->Tx_B, mh->Ty_B);
					dim3 grid (mh->B_D);
					cudaFuncSetCacheConfig (MUBLAS_GEMV_T_DEVICE_v2, cudaFuncCachePreferL1);
					if (mh->cc >= 300) MUBLAS_GEMV_T_DEVICE_v2 <<< grid, threads >>> (m, n, *alpha, a, lda, x, *beta, y);
					else MUBLAS_GEMV_T_DEVICE_v2 <<< grid, threads, mh->S_B >>> (m, n, *alpha, a, lda, x, *beta, y);
				} else { // for misaligned, scalar version
					mh->kernelnumber = 210;
					cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_T_DEVICE_s1);
					get_Param_gemvt_2 (mh);
					dim3 threads (mh->Tx_B, mh->Ty_B);
					dim3 grid (mh->B_D);
					cudaFuncSetCacheConfig (MUBLAS_GEMV_T_DEVICE_s1, cudaFuncCachePreferL1);
					if (mh->cc >= 300) MUBLAS_GEMV_T_DEVICE_s1 <<< grid, threads >>> (m, n, *alpha, a, lda, x, *beta, y);
					else MUBLAS_GEMV_T_DEVICE_s1 <<< grid, threads, mh->S_B >>> (m, n, *alpha, a, lda, x, *beta, y);
				}
			} else { // scalar version
				mh->kernelnumber = 210;
				cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_T_DEVICE_s1);
				get_Param_gemvt_2 (mh);
				dim3 threads (mh->Tx_B, mh->Ty_B);
				dim3 grid (mh->B_D);
				cudaFuncSetCacheConfig (MUBLAS_GEMV_T_DEVICE_s1, cudaFuncCachePreferL1);
				if (mh->cc >= 300) MUBLAS_GEMV_T_DEVICE_s1 <<< grid, threads >>> (m, n, *alpha, a, lda, x, *beta, y);
				else MUBLAS_GEMV_T_DEVICE_s1 <<< grid, threads, mh->S_B >>> (m, n, *alpha, a, lda, x, *beta, y);
			}
		}
		else { // INCX,INCY != 1
			#if defined (PREC_S) || defined (PREC_C) || defined (PREC_D) || defined (PREC_Z)
			mh->kernelnumber = 8200;
			cublasHandle_t ch;
			cublasCreate (&ch);
			cublasSetPointerMode(ch, CUBLAS_POINTER_MODE_HOST);
			cublasOperation_t opAcublas = CUBLAS_OP_T;
			CUBLAS_GEMV (ch, opAcublas, m, n, alpha, a, lda, x, incx, beta, y, incy);
			#else
			printf ("mublas error: not implemented");
			#endif
		}
	}
	return 0;
}


