#ifndef PREC_W_H
#define PREC_W_H

#include "cudd.h"

#ifdef PREC_W
	// host
	#define MUBLAS_GEMM mublasWgemm
	#define MUBLAS_GEMV mublasWgemv
	#define MUBLAS_TRMV mublasWtrmv
	#define MUBLAS_AXPY mublasWaxpy
	#define MUBLAS_DOT  mublasWdot
	#define MUBLAS_SCAL mublasWscal
	#define MUBLAS_COPY mublasWcopy
//	#define CUBLAS_GEMM
//	#define CUBLAS_GEMV
//	#define CUBLAS_TRMV 
//	#define CUBLAS_AXPY
//	#define CUBLAS_DOT
//	#define CUBLAS_SCAL
//	#define CUBLAS_COPY
	#define TYPE0 cuddreal
	#define ZERO0 make_cuddreal(0.,0.)
	#define ONE0 make_cuddreal(1.,0.)

	// vec1
	#define VEC1 1
	#define TYPE1 TYPE0
	#define ZERO1 ZERO0
	#define ADD1 CUWADD
	#define MUL1 CUWMUL
	#define MAD1 CUWMAD

	#define TYPEIN1 TYPE1
	#define ZEROIN1 ZERO1
	#define LOAD1___ ld_global_v2_f64
	#define LOAD1_NC ld_global_nc_v2_f64

	// vec2
	#define VEC2 0

	// vec4
	#define VEC4 0

	// DD-algorithm 
	#define CUWADD cuWaddV2
	#define CUWSUB cuWsubV2
	#define CUWMUL cuWmulV3
	#define CUWMAD cuWmuladd
#endif

// === Shuffle =====================================================
__device__ __forceinline__ void shuffleAdd (cuddreal &val, const int n) {
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
	val = CUWADD (val, make_cuddreal ( 
		__hiloint2double( __shfl_xor(__double2hiint(val.x), n, T_W), __shfl_xor(__double2loint(val.x), n, T_W)),
		__hiloint2double( __shfl_xor(__double2hiint(val.y), n, T_W), __shfl_xor(__double2loint(val.y), n, T_W))));
#endif
}
// =================================================================

// === Load & Store ================================================
__device__ __forceinline__ void ld_global_v2_f64 (cuddreal &reg, const cuddreal* addr) {
#if USE_INLINE_ASM
	asm ("ld.global.v2.f64 {%0,%1}, [%2+0];": "=d"(reg.x), "=d"(reg.y): "l"(addr));
#else
	reg = *addr;
#endif
}

__device__ __forceinline__ void ld_global_nc_v2_f64 (cuddreal &reg, const cuddreal* addr) {
#if USE_INLINE_ASM
	#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 320)
	asm ("ld.global.nc.v2.f64 {%0,%1}, [%2+0];": "=d"(reg.x), "=d"(reg.y): "l"(addr));
	#else
	asm ("ld.global.v2.f64 {%0,%1}, [%2+0];": "=d"(reg.x), "=d"(reg.y): "l"(addr));
	#endif
#else
	reg = *addr;
#endif
}
// =================================================================

__host__ __device__ __forceinline__ bool operator== (const cuddreal a, const cuddreal b) {
	if ((a.x == b.x) && (a.y == b.y))
		return true;
	else
		return false;
}

__device__ __forceinline__ cuddreal
cuWmuladd
	(const cuddreal a, const cuddreal b, const cuddreal c)
{
	return CUWADD(CUWMUL(a, b), c);
}

#endif
