#ifndef TESTING_COMMON
#define TESTING_COMMON

// header ------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#define __STDC_FORMAT_MACROS
#include <inttypes.h>
#include <sys/time.h>
#include <unistd.h>
#ifndef QD_CHECK
#include <f2c.h>
#endif
#include <algorithm>
#include <cuda.h>
#include <cuda_runtime_api.h>
#include <math.h>
#include <cblas.h>
#include <cublas_v2.h>
#include <getopt.h>
#include <ctype.h>

#ifdef USE_NVML
#include <nvml.h>
#endif

#ifdef QD_CHECK
#include <qd/qd_real.h>
#include <mblas_qd.h>
#define REFERENCE_TYPE MBLAS_TYPE
#define REFERENCE_BASETYPE MBLAS_BASETYPE
#define REFERENCE_ZERO MBLAS_ZERO
#define REFERENCE_ONE MBLAS_ONE
#define REFERENCE_NEGONE MBLAS_NEGONE
#define REFERENCE_AXPY MBLAS_AXPY
#define REFERENCE_NRM MBLAS_NRM
#else
#include <clapack.h>
#define REFERENCE_TYPE CBLAS_TYPE
#define REFERENCE_BASETYPE CBLAS_BASETYPE
#define REFERENCE_ZERO CBLAS_ZERO
#define REFERENCE_ONE CBLAS_ONE
#define REFERENCE_NEGONE CBLAS_NEGONE
#define REFERENCE_AXPY CBLAS_AXPY
#define REFERENCE_NRM CBLAS_NRM
#endif

#if defined MUBLAS
#include <mublas.h>
#elif defined CUBLAS
#elif defined KBLAS
#include <kblas.h> 
#elif defined MAGMA
#include <magma.h>
#elif defined CULA
#include <cula_blas_device.h>
#elif defined ASPEN
#include <aspen.h>
#endif

#ifdef PREC_S
#include "prec_s.h"
#endif
#ifdef PREC_D
#include "prec_d.h"
#endif
#ifdef PREC_C
#include "prec_c.h"
#endif
#ifdef PREC_Z
#include "prec_z.h"
#endif
#ifdef PREC_W
#include "prec_w.h"
#endif
#ifdef PREC_U
#include "prec_u.h"
#endif
#ifdef PREC_B
#include "prec_b.h"
#endif
#ifdef PREC_V
#include "prec_v.h"
#endif

#if defined GEMM
#define ROUTINE_FLOPS(m,n,k)			(FPMADCOUNT*m*n*k+(2*FPMULCOUNT+FPADDCOUNT)*m*n)
#define ROUTINE_BYTES(m,n,k,s)			((m*k+k*n+2*m*n)*s) 
#define ROUTINE_MAXSIZE(gmem,margin,s)	(sqrt((gmem-margin)/s/3)) 
#define ROUTINE_MAXSIZE_1(gmem,margin,s,dim1)		((gmem-margin)/s/dim1) // !!!!!!!!!!!!!!
#define ROUTINE_MAXSIZE_2(gmem,margin,s,dim1,dim2)	((gmem+margin+s+dim1+dim2)*0)  // !!!!!!!!!!!!!!!
#elif defined GEMV
#define ROUTINE_FLOPS(m,n,k)						(FPMADCOUNT*m*n+2*FPMULCOUNT*m+FPADDCOUNT*m+(k*0))
#define ROUTINE_BYTES(m,n,k,s)						((2*m+n+m*n)*s+(k*0)) 
#define ROUTINE_MAXSIZE(gmem,margin,s)				(sqrt((gmem-margin)/s))
#define ROUTINE_MAXSIZE_1(gmem,margin,s,dim1)		((gmem-margin)/s/dim1)
#define ROUTINE_MAXSIZE_2(gmem,margin,s,dim1,dim2)	((gmem+margin+s+dim1+dim2)*0) 
#elif defined TRMV
#define ROUTINE_FLOPS(m,n,k)						(0)
#define ROUTINE_BYTES(m,n,k,s)						trmv_mem(m)*s+(n*k*0) 
#define ROUTINE_MAXSIZE(gmem,margin,s)				(sqrt((gmem-margin)/s))
#define ROUTINE_MAXSIZE_1(gmem,margin,s,dim1)		((gmem-margin)/s/dim1)
#define ROUTINE_MAXSIZE_2(gmem,margin,s,dim1,dim2)	((gmem+margin+s+dim1+dim2)*0) 
#elif defined SYMV
#define ROUTINE_FLOPS(m,n,k)						(FPMADCOUNT*n*n+2*FPMULCOUNT*n+FPADDCOUNT*n+(m*k*0))
#define ROUTINE_BYTES(m,n,k,s)						symv_mem(n)*s+(m*k*0) //((3*n+n*n)*s+(m*k*0)) 
#define ROUTINE_MAXSIZE(gmem,margin,s)				(sqrt((gmem-margin)/s))
#define ROUTINE_MAXSIZE_1(gmem,margin,s,dim1)		((gmem-margin)/s/dim1)
#define ROUTINE_MAXSIZE_2(gmem,margin,s,dim1,dim2)	((gmem+margin+s+dim1+dim2)*0) 
#elif defined HEMV
#define ROUTINE_FLOPS(m,n,k)						(FPMADCOUNT*n*n+2*FPMULCOUNT*n+FPADDCOUNT*n+(m*k*0))
#define ROUTINE_BYTES(m,n,k,s)						symv_mem(n)*s+(m*k*0) //((3*n+n*n)*s+(m*k*0)) 
#define ROUTINE_MAXSIZE(gmem,margin,s)				(sqrt((gmem-margin)/s))
#define ROUTINE_MAXSIZE_1(gmem,margin,s,dim1)		((gmem-margin)/s/dim1)
#define ROUTINE_MAXSIZE_2(gmem,margin,s,dim1,dim2)	((gmem+margin+s+dim1+dim2)*0) 
#elif defined TRSV
#define ROUTINE_FLOPS(m,n,k)						(0)
#define ROUTINE_BYTES(m,n,k,s)						trmv_mem(m)*s+(n*k*0) 
#define ROUTINE_MAXSIZE(gmem,margin,s)				(sqrt((gmem-margin)/s))
#define ROUTINE_MAXSIZE_1(gmem,margin,s,dim1)		((gmem-margin)/s/dim1)
#define ROUTINE_MAXSIZE_2(gmem,margin,s,dim1,dim2)	((gmem+margin+s+dim1+dim2)*0) 
#elif defined GER
#define ROUTINE_FLOPS(m,n,k)						(FPMADCOUNT*m*n+2*FPMULCOUNT*m*n+FPADDCOUNT*m*n+(k*0))
#define ROUTINE_BYTES(m,n,k,s)						((m+n+2*m*n)*s+(k*0)) 
#define ROUTINE_MAXSIZE(gmem,margin,s)				(sqrt((gmem-margin)/s))
#define ROUTINE_MAXSIZE_1(gmem,margin,s,dim1)		((gmem-margin)/s/dim1)
#define ROUTINE_MAXSIZE_2(gmem,margin,s,dim1,dim2)	((gmem+margin+s+dim1+dim2)*0) 
#endif
#if defined AXPY
#define ROUTINE_FLOPS(m,n,k)			(FPMADCOUNT*n+(k+m)*0)
#define ROUTINE_BYTES(m,n,k,s)			(3*n*s+(k+m)*0)
#define ROUTINE_MAXSIZE(gmem,margin,s)	((gmem-margin)/2/s)
#define ROUTINE_MAXSIZE_1(gmem,margin,s,dim1)		(gmem+margin+s+dim1)*0
#define ROUTINE_MAXSIZE_2(gmem,margin,s,dim1,dim2)	(gmem+margin+s+dim1+dim2)*0 
#elif defined DOT
#define ROUTINE_FLOPS(m,n,k)			(FPMADCOUNT*n+(k+m)*0)
#define ROUTINE_BYTES(m,n,k,s)			(2*n*s+(k+m)*0)
#define ROUTINE_MAXSIZE(gmem,margin,s)	((gmem-margin)/2/s)
#define ROUTINE_MAXSIZE_1(gmem,margin,s,dim1)		(gmem+margin+s+dim1)*0
#define ROUTINE_MAXSIZE_2(gmem,margin,s,dim1,dim2)	(gmem+margin+s+dim1+dim2)*0 
#elif defined SCAL
#define ROUTINE_FLOPS(m,n,k)			(FPMULCOUNT*n+(k+m)*0)
#define ROUTINE_BYTES(m,n,k,s)			(2*n*s+(m+k)*0)
#define ROUTINE_MAXSIZE(gmem,margin,s)	((gmem-margin)/s)
#define ROUTINE_MAXSIZE_1(gmem,margin,s,dim1)		(gmem+margin+s+dim1)*0
#define ROUTINE_MAXSIZE_2(gmem,margin,s,dim1,dim2)	(gmem+margin+s+dim1+dim2)*0 
#elif defined COPY
#define ROUTINE_FLOPS(m,n,k)			((m+n+k)*0)
#define ROUTINE_BYTES(m,n,k,s)			(2*n*s+(m+k)*0)
#define ROUTINE_MAXSIZE(gmem,margin,s)	((gmem-margin)/s)
#define ROUTINE_MAXSIZE_1(gmem,margin,s,dim1)		(gmem+margin+s+dim1)*0
#define ROUTINE_MAXSIZE_2(gmem,margin,s,dim1,dim2)	(gmem+margin+s+dim1+dim2)*0 
#endif

int64_t trmv_mem (int64_t n) {
	int64_t i, v = 0;
	for (i = 0; i < n; i++) 
		v += (n - i);
	v += 2 * n;
	return v;
}

int64_t symv_mem (int64_t n) {
	int64_t i, v = 0;
	for (i = 0; i < n; i++) 
		v += (n - i);
	v += 3 * n;
	return v;
}

#include "testing_setting.h"

#endif
