#include "testing_common.cpp"

// routine ------------------------------
//
#ifdef HEMV
// target routine
#ifdef MUBLAS
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y)
#elif CUBLAS
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) CUBLAS_HEMV(ch,cublas_uplo_const_m(th.uplo),n,&al,A,lda,X,th.incx,&bt,Y,th.incy)
#elif KBLAS
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) KBLAS_HEMV(th.uplo,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#elif MAGMA
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) MAGMA_HEMV(magma_uplo_const(th.uplo),n,al,A,lda,X,th.incx,bt,Y,th.incy)
#elif CULA
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) CULA_HEMV(th.uplo,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#elif ASPEN
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) ASPEN_HEMV(th.uplo,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#endif
// reference routine
#ifdef QD_CHECK
#define ROUTINE_REFERENCE(th,m,n,al,A,lda,X,bt,Y) MBLAS_HEMV(&th.uplo,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#else
#define ROUTINE_REFERENCE(th,m,n,al,A,lda,X,bt,Y) CBLAS_HEMV(CblasColMajor,cblas_uplo_const_m(th.uplo),n,&al,A,lda,X,th.incx,&bt,Y,th.incy)
#endif
#endif

#ifdef SYMV
// target routine
#ifdef MUBLAS
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y)
#elif CUBLAS
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) CUBLAS_SYMV(ch,cublas_uplo_const_m(th.uplo),n,&al,A,lda,X,th.incx,&bt,Y,th.incy)
#elif KBLAS
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) KBLAS_SYMV(th.uplo,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#elif MAGMA
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) MAGMA_SYMV(magma_uplo_const(th.uplo),n,al,A,lda,X,th.incx,bt,Y,th.incy)
#elif CULA
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) CULA_SYMV(th.uplo,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#elif ASPEN
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) ASPEN_SYMV(th.uplo,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#endif
// reference routine
#ifdef QD_CHECK
#define ROUTINE_REFERENCE(th,m,n,al,A,lda,X,bt,Y) MBLAS_SYMV(&th.uplo,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#else
#define ROUTINE_REFERENCE(th,m,n,al,A,lda,X,bt,Y) CBLAS_SYMV(CblasColMajor,cblas_uplo_const_m(th.uplo),n,al,A,lda,X,th.incx,bt,Y,th.incy)
#endif
#endif

#ifdef GEMV
// target routine
#ifdef MUBLAS
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) MUBLAS_GEMV(&mh,th.tranA,m,n,&al,A,lda,X,th.incx,&bt,Y,th.incy)
#elif CUBLAS
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) CUBLAS_GEMV(ch,cublas_trans_const_m(th.tranA),m,n,&al,A,lda,X,th.incx,&bt,Y,th.incy)
#elif KBLAS
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) KBLAS_GEMV(th.tranA,m,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#elif MAGMA
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) MAGMA_GEMV(magma_trans_const(th.tranA),m,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#elif CULA
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) CULA_GEMV(th.tranA,m,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#elif ASPEN
#define ROUTINE_TARGET(th,m,n,al,A,lda,X,bt,Y) ASPEN_GEMV(th.tranA,m,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#endif
// reference routine
#ifdef QD_CHECK
#define ROUTINE_REFERENCE(th,m,n,al,A,lda,X,bt,Y) MBLAS_GEMV(&th.tranA,m,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#else
#if defined (PREC_S) || defined (PREC_D) || defined (PREC_W) || defined (PREC_B)
#define ROUTINE_REFERENCE(th,m,n,al,A,lda,X,bt,Y) CBLAS_GEMV(CblasColMajor,cblas_trans_const_m(th.tranA),m,n,al,A,lda,X,th.incx,bt,Y,th.incy)
#else
#define ROUTINE_REFERENCE(th,m,n,al,A,lda,X,bt,Y) CBLAS_GEMV(CblasColMajor,cblas_trans_const_m(th.tranA),m,n,&al,A,lda,X,th.incx,&bt,Y,th.incy)
#endif
#endif
#endif

// routine ------------------------------


int32_t
main (int32_t argc, char **argv)
{
	int64_t i, dim_mm, dim_nn, lda, lda_cpu, iter = MINLOOP;
	int64_t sizem = sizeof (REFERENCE_TYPE);
	float sec;
	GPU_TYPE alpha, beta;
	GPU_TYPE *hst_A, *hst_X, *hst_Y;
	GPU_TYPE *hst_Y_t;
	GPU_TYPE *dev_A, *dev_X, *dev_Y;
	REFERENCE_TYPE *hst_A_r = 0, *hst_X_r = 0, *hst_Y_r = 0, alpha_r = REFERENCE_ZERO, beta_r = REFERENCE_ZERO;
// testing setup ------------------------------
	testingHandle_t th;
	testingCreate (&th);
	th.word_size = sizeof(GPU_TYPE);
	set_testing_handler (argc, argv, &th);
	dim_initialize (&th);
	print_info (&th);
	print_info2 (&th);
// testing setup ------------------------------
// library setup ------------------------------
	cudaEvent_t t_start, t_stop;
	cudaEventCreate (&t_start);
	cudaEventCreate (&t_stop);
	cudaError_t err;

	#ifdef USE_NVML
	uint32_t temp1 = 0, temp2 = 0;
	nvmlDevice_t dev;
	nvmlInit();
	nvmlDeviceGetHandleByIndex(DEVNO, &dev);
	do {
		sleep(1);
		nvmlDeviceGetTemperature (dev, NVML_TEMPERATURE_GPU, &temp1);
	} while (temp1 >= TEMPLIMIT);
	#endif

	#ifdef MUBLAS 
	mublasHandle_t mh;
	mublasCreate (&mh);
	set_mublas_handler (&th, &mh);
	#elif CUBLAS
	cublasHandle_t ch;
	cublasCreate (&ch);
	cublasSetPointerMode(ch, CUBLAS_POINTER_MODE_HOST);
	if (th.cublasatomic == 1) 
		cublasSetAtomicsMode(ch, CUBLAS_ATOMICS_ALLOWED);
	else
		cublasSetAtomicsMode(ch, CUBLAS_ATOMICS_NOT_ALLOWED);
	#elif KBLAS
	#elif MAGMA
	magma_init();
	#elif CULA
    culaInitialize();
	#elif ASPEN
	ASPEN_init(0);
	#else
	#endif
#if DEBUG_ALLSKIP == 0
// library setup ------------------------------
// memory setup -------------------------------
	cudaMallocHost ((void **) &hst_A, th.word_size * th.dim_m * th.dim_n);
	cudaMallocHost ((void **) &hst_X, th.word_size * th.dim_n * th.incx);
	cudaMallocHost ((void **) &hst_Y, th.word_size * th.dim_m * th.incy);
	init_vector (1, &alpha);
	init_vector (1, &beta);
	init_vector (th.dim_m * th.dim_n, hst_A);
// Hermitian (for MAGMA)
	for (i=0; i<th.dim_m; i++) {
#if defined (PREC_C)
		hst_A[i*th.dim_m+i] = make_cuFloatComplex(hst_A[i*th.dim_m+i].x,0);
#elif defined (PREC_Z)
		hst_A[i*th.dim_m+i] = make_cuDoubleComplex(hst_A[i*th.dim_m+i].x,0);
#endif
	}
	init_vector (th.dim_n * th.incx, hst_X);
	init_vector (th.dim_m * th.incy, hst_Y);
	if (th.mode == 'c') {
		cudaMallocHost ((void **) &hst_Y_t, th.word_size * th.dim_m * th.incy);
		hst_A_r = (REFERENCE_TYPE *) malloc (sizem * th.dim_m * th.dim_n);
		hst_X_r = (REFERENCE_TYPE *) malloc (sizem * th.dim_n * th.incx);
		hst_Y_r = (REFERENCE_TYPE *) malloc (sizem * th.dim_m * th.incy);
		copy_vector (th.dim_m * th.dim_n, hst_A, hst_A_r);
		copy_vector (th.dim_n * th.incx, hst_X, hst_X_r);
		copy_vector (1, &alpha, &alpha_r);
		copy_vector (1, &beta, &beta_r);
	}

// memory setup -------------------------------
	printf ("%% ==========================================\n");
	printf ("%% M\tN");
	if (th.mode == 'p') printf ("\tTime[sec]\t[GFlops]\t[GB/s]\t(#iter)");
	if (th.mode == 'c') printf ("\tERROR");
	#ifdef MUBLAS
	printf ("\tTx_B\tTy_B\tS_B\tB_D\tB_M_max\tBO\tWO\tBOWO\tkernel#");
	#endif
	#ifdef USE_NVML
	printf ("\tTemp.bf[C]\tTemp.af[C]");
	#endif
	printf ("\n");
// evaluation ---------------------------------
	dim_start (&th);
	while (1) {
		if ((th.const_m == 0 && th.dim_m > th.dim_stop) ||
		    (th.const_n == 0 && th.dim_n > th.dim_stop) ||
		    (th.const_k == 0 && th.dim_k > th.dim_stop)) break;

		if (th.tranA == 'T' || th.tranA == 't' || th.tranA == 'C' || th.tranA == 'c') {
			lda = th.dim_n;
			lda_cpu = th.dim_n;
			dim_mm = th.dim_n;
			dim_nn = th.dim_m;
		} else {
			lda = th.dim_m;
			if (th.const_m == 1) lda_cpu = th.dim_m;
			else lda_cpu = th.dim_stop;
			dim_mm = th.dim_m;
			dim_nn = th.dim_n;
		}

		#if PITCHALLOC
		size_t pitch;
		cudaMallocPitch ((void **) &dev_A, &pitch, th.word_size * lda, dim_nn);
		lda = pitch/th.word_size;
		#else
		cudaMalloc ((void **) &dev_A, th.word_size * dim_mm * dim_nn);
		#endif
		cudaMalloc ((void **) &dev_X, th.word_size * th.dim_n * th.incx);
		cudaMalloc ((void **) &dev_Y, th.word_size * th.dim_m * th.incy);

		cublasSetMatrix(dim_mm, dim_nn, th.word_size, hst_A, lda_cpu, dev_A, lda);
		cublasSetVector(th.dim_n * th.incx, th.word_size, hst_X, 1, dev_X, 1);
		cublasSetVector(th.dim_m * th.incy, th.word_size, hst_Y, 1, dev_Y, 1);

		#ifdef USE_NVML
		nvmlDeviceGetTemperature (dev, NVML_TEMPERATURE_GPU, &temp1);
		#endif

		printf ("%zd\t%zd", dim_mm, dim_nn);
		get_routine_theoretial_performance (&th);

		// execution ---------------------------------
		if (th.mode == 'p') {
			#if PERF_ACCURATE
			iter = std::min ((double)MAXLOOP, std::max ((double)MINLOOP, MINSEC / std::max
				(th.routine_flops / th.routine_flops_per_sec_theoretical_peak, th.routine_bytes / th.device_bytes_per_sec)));
			for (i = 0; i < 2; i++) ROUTINE_TARGET(th,dim_mm,dim_nn,alpha,dev_A,lda,dev_X,beta,dev_Y); // dummy
			cudaEventRecord (t_start, 0);
			for (i = 0; i < iter; i++) ROUTINE_TARGET(th,dim_mm,dim_nn,alpha,dev_A,lda,dev_X,beta,dev_Y);
			cudaEventRecord (t_stop, 0);
			cudaEventSynchronize (t_stop);
			cudaEventElapsedTime (&sec, t_start, t_stop);
			sec = 1.0e-3 * sec / (float)iter;
			#else
			float sec_all_exec[MINLOOP];
			iter = MINLOOP;
			for (i = 0; i < iter; i++) {
				cudaEventRecord (t_start, 0);
				ROUTINE_TARGET(th,dim_mm,dim_nn,alpha,dev_A,lda,dev_X,beta,dev_Y);
				cudaEventRecord (t_stop, 0);
				cudaEventSynchronize (t_stop);
				cudaEventElapsedTime (&sec, t_start, t_stop);
				sec_all_exec[i] = sec;
			}
			sec = 0.;
			for (i = 0; i < iter; i++) {
				if (sec_all_exec[i] > sec) sec = sec_all_exec[i];
			}
			sec = 1.0e-3 * sec;
			#endif
		}
		if (th.mode == 'c') {
			ROUTINE_TARGET(th,dim_mm,dim_nn,alpha,dev_A,lda,dev_X,beta,dev_Y);
		}
		// execution ---------------------------------

		err = cudaGetLastError();
		if (err != cudaSuccess) printf("error : %s\n", cudaGetErrorString(err));
		#ifdef USE_NVML
		nvmlDeviceGetTemperature (dev, NVML_TEMPERATURE_GPU, &temp2);
		#endif

		if (th.mode == 'p') {
			printf ("\t%1.3e\t%1.3e\t%1.3e\t(%zd)", sec, (th.routine_flops / sec) * 1.0e-9, (th.routine_bytes / sec) * 1.0e-9, iter);
		}
		if (th.mode == 'c') {
			cublasGetVector (th.dim_m * th.incy, th.word_size, dev_Y, 1, hst_Y_t, 1);
			copy_vector (th.dim_m * th.incy, hst_Y, hst_Y_r);
			ROUTINE_REFERENCE(th,dim_mm,dim_nn,alpha_r,hst_A_r,lda_cpu,hst_X_r,beta_r,hst_Y_r);
			check_result (th.dim_m * th.incy, hst_Y_t, hst_Y_r);
		}

		#ifdef MUBLAS
		print_mublas_kernel_info (&mh);
		#endif
		#ifdef USE_NVML
		printf ("\t%10u\t%10u", temp1, temp2);
		#endif
		printf ("\n");

		cudaFree (dev_A);
		cudaFree (dev_X);
		cudaFree (dev_Y);

		dim_step (&th);
	}
	printf ("%% ==========================================\n");
// evaluation ---------------------------------
// shutdown -----------------------------------
	cudaFreeHost (hst_A);
	cudaFreeHost (hst_X);
	cudaFreeHost (hst_Y);
	if (th.mode == 'c') {
		free (hst_A_r);
		free (hst_X_r);
		free (hst_Y_r);
		cudaFreeHost (hst_Y_t);
	}

	#ifdef MUBLAS
	print_info3 (&mh);
	#elif CUBLAS
	cublasDestroy (ch);
	#elif KBLAS
	#elif MAGMA
	#elif CULA
	#elif ASPEN
	ASPEN_shutdown();
	#endif
	#ifdef USE_NVML
	nvmlShutdown();
	#endif
	cudaThreadExit();
// shutdown -----------------------------------

#endif
	return 0;
}

